\input texinfo   @c -*-texinfo-*-

@c %**start of header
@setfilename report.info
@settitle A Uniformity Test for Multidimensional Data
@c %**end of header

@setchapternewpage odd

@c @libavl{} macro
@iftex
@macro libavl{}
@sc{Libavl}
@end macro
@end iftex
@ifnottex
@macro libavl{}
@code{libavl}
@end macro
@end ifnottex

@c @bibref{} macro
@iftex
@macro bibref{cite}
[\cite\]
@end macro
@end iftex
@ifinfo
@macro bibref{cite}
@ref{\cite\}
@end macro
@end ifinfo
@ifhtml
@macro bibref{cite}
[@ref{\cite\}]
@end macro
@end ifhtml

@c References clauses
@ifinfo
@macro references
@noindent References:@w{ }
@end macro
@end ifinfo
@ifnotinfo
@macro references
@noindent @strong{References:}@w{ }
@end macro
@end ifnotinfo
@macro bibdfn{cite}
@noindent @anchor{\cite\}
[\cite\].@w{  }
@end macro

@macro WEB{}
@t{WEB}
@end macro

@macro TexiWEB{}
TexiWEB
@end macro

@c METAFONT logo
@tex
\global\font\logo=logo10
\global\def\MF{\strut\hbox{\logo METAFONT}}\def\.#1{\strut\hbox{\tt #1}}
@end tex
@ifnottex
@macro MF{}
METAFONT
@end macro
@end ifnottex

@c I don't like the default Info style for book titles: `Title'.
@c In my opinion, _Title_ looks much nicer.
@definfoenclose cite,_,_

@c Using italics for definitions makes them look like identifiers.
@iftex
@alias dfn=b
@end iftex

@c Centered dots for typesetting relations.
@ifnottex
@alias cdots=dots
@end ifnottex

@c Glossary references
@iftex
@macro gloss{term}
@dfn{\term\}
@end macro
@end iftex
@ifnottex
@macro gloss{term}
@dfn{\term\} (@pxref{\term\})
@end macro
@end ifnottex
@macro glossdfn{term}
@anchor{\term\}
@dfn{\term\}:
@end macro

@c Nice mathematics or ASCII mathematics, depending on format
@iftex
@macro altmath{TEX, ASCII}
@tex
$\TEX\$@unskip
@end tex
@end macro
@end iftex
@ifnottex

@macro altmath{TEX, ASCII}
@math{\ASCII\}
@end macro
@end ifnottex
@macro comma
,
@end macro
@twheader

@titlepage
@title Uniformity Testing
@subtitle Testing for uniformity in multidimensional data
@subtitle with the MST-based test of Smith and Jain
@author Ben Pfaff 

@page
@vskip 0pt plus 1filll

Except where noted, this work falls under this license:

@quotation
Copyright @copyright{} 2001 Ben Pfaff.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

@itemize @bullet
@item
Redistributions of source code must retain the above copyright notice,
this list of conditions, and the following disclaimer.

@item
Redistributions in binary form must reproduce the above copyright
notice, this list of conditions, and the following disclaimer in the
documentation and/or other materials provided with the distribution.
@end itemize

There is NO WARRANTY, not even for MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.
@end quotation

Some parts of this work are subject to this license:

@quotation
Copyright @copyright{} 2001 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

@display
Free Software Foundation, Inc.
59 Temple Place - Suite 330
Boston, MA 02111-1307
UNITED STATES
@end display
@end quotation

The author may be contacted as @email{blp@@gnu.org} on the Internet, or
write to:

@display
Ben Pfaff
12167 Airport
DeWitt, MI 48820
UNITED STATES
@end display
@end titlepage

@contents

@ifnottex
@node Top, Introduction, (dir), (dir)
@top Uniformity Testing

@menu
* Introduction::                
* Uniformity Testing::          
* Minimum Spanning Tree Computation::  
* Priority Queues::             
* Other Classes::               
* Discussion::                  
* References::                  
* Index::                       

@detailmenu
 --- The Detailed Node Listing ---

Introduction

* How to Read a TexiWEB::       
* Coding Style::                
* License::                     

Uniformity Testing

* External Interface::          
* Implementation::              

External Interface: @file{uniformity.h}

* Low-Level Routines::          
* Options::                     

Implementation: @file{uniformity.c}

* unf_test() implementation::   
* unf_inside_hull() implementation::  
* unf_run_mst() implementation::  
* unf_calc_results() implementation::  
* unf_init_options() implementation::  

Minimum Spanning Tree Computation

* MST Class::                   
* The Algorithm of Prim::       
* MST Class Implementations::   

Priority Queues

* PQ Class::                    
* Binary Heap PQ::              
* Fibonacci Heap PQ::           

Binary Heap

* Binary heap creation::        
* Binary heap destruction::     
* Binary heap extraction::      
* Binary heap decrease key::    
* Binary heap count::           
* Binary heap key lookup::      
* Binary heap verification::    
* Binary heap dump::            
* Binary heap class implementation::  

Fibonacci Heap

* Fibonacci Heap Creation::     
* Fibonacci Heap Destruction::  
* Fibonacci Heap Extract Minimum::  
* Fibonacci Heap Decrease Key::  
* Fibonacci Heap Count::        
* Fibonacci Heap Key Lookup::   
* Fibonacci Heap Verification::  
* Fibonacci Heap Dumping::      
* Fibonacci Heap Class Implementation::  

Other Classes

* Memory Allocation::           
* Random Number Generator::     
* Point Set::                   

Discussion

* Programs for Testing::        
* Testing Results::             
* Sources of Error::            
* Performance::                 
* Future Work::                 

Programs for Testing

* Testing Convex Hull Estimation::  
* Testing Minimum Spanning Trees::  
* Testing the Uniformity Tester::  

Testing Results

* Correctness::                 
* Normal Distributions::        
* Testing Separation::          

Correctness

* Known Sampling Window::       
* Unknown Sampling Window::     

Sources of Error

* Bugs in MST Implementation::  
* Bugs in Convex Hull Insideness Test::  
* Imprecise Arithmetic::        

@end detailmenu
@end menu

@end ifnottex
@include intro.w
@include uniformity.w
@include mst.w
@include pq.w
@include odds.w
@include test.w
@include references.w
@include index.w

@bye
