@c -*-texinfo-*-

@node Discussion, References, Other Classes, Top
@chapter Discussion

This chapter discusses the library for uniformity testing.  It begins
with a description of the provided programs for testing.  Sample results
from these programs are then displayed and discussed.  Next,
implications for the correctness of the library and its components are
suggested.  The library's performance is analyzed.  Finally, some
directions for further work are mentioned.

@menu
* Programs for Testing::        
* Testing Results::             
* Sources of Error::            
* Performance::                 
* Future Work::                 
@end menu

@node Programs for Testing, Testing Results, Discussion, Discussion
@section Programs for Testing

This section briefly discusses the programs for testing the uniformity
testing library as a whole or in part.  The source code for these
programs is not presented due to its length and tedium.

@menu
* Testing Convex Hull Estimation::  
* Testing Minimum Spanning Trees::  
* Testing the Uniformity Tester::  
@end menu

@node Testing Convex Hull Estimation, Testing Minimum Spanning Trees, Programs for Testing, Programs for Testing
@subsection Convex Hull Estimation

The first step in the uniformity testing algorithm is to select a random
set of points within the approximate convex hull of the points to be
tested.  It is worthwhile to test that these points are generated
properly.  

The @code{hull-test} program tests the convex hull estimation visually.
It picks an initial set of points within a triangular area, then
generates another set in their approximate convex hull using
|unf_inside_hull()|.  It prints a graphical representation of both sets
of points to |stdout| in a format suitable for the @code{jgraph}
graphing utility.  The user can then judge for himself whether the
points were correctly generated.

@node Testing Minimum Spanning Trees, Testing the Uniformity Tester, Testing Convex Hull Estimation, Programs for Testing
@subsection Minimum Spanning Trees

The code for priority queues for the Prim MST is somewhat complex, and
therefore error-prone.  A simple test to make sure that the priority
queues are probably implemented correctly is helpful.  This is what the
@code{pq-test} program does.  

When run without command-line arguments, @code{pq-test} will perform its
test 16 times for a priority queue of 128 items and print a success or
failure message for each test.  To see a list of other options, use the
command @samp{pq-test -h}.

The minimum spanning tree algorithm as a whole is also worth testing.
The @code{mst-test} program does this.  It works similarly to the
@code{hull-test} program, generating a random set of points, calculating
their MST, and producing as output a graphical representation of the
MST.

@node Testing the Uniformity Tester,  , Testing Minimum Spanning Trees, Programs for Testing
@subsection Uniformity Tester

Testing the uniformity testing library itself is important, too.  The
@code{unf-test} program does this.  Each time it is run it generates a
set of points randomly within user-specified parameters, then runs a
uniformity test on it and reports the results.  The user can specify
several parameters for the test using command-line arguments:

@table @code
@item -d @var{dims}
@itemx --dims=@var{dims}
|d|, the number of dimensions.

@item -c @var{count}
@itemx --count=@var{count}
|n|, the number of points to test.

@item -a @var{avg}
@itemx --avg=@var{avg}
Optionally, a number of calls to |unf_test()| over which to average
|s|.  By default only one call is used per data set.

@item -S @var{sig}
@itemx --sig=@var{sig}
The minimum value of |s| that is considered ``uniform,'' .05 by default.
Increasing this value will decrease the chance of non-uniform data sets
considered uniform and increase the chance of uniform data sets
considered non-uniform.

@item -D @var{dist}
@itemx --dist=@var{dist}
The distribution from which the test points are picked.  Argument
@var{dist} may be one of the following:

@table @code
@item uniform
The uniform distribution on [0, 1].

@item decreasing
The distribution of decreasing values such that @altmath{x_1 > x_2 >
\dots{} > x_d, X1 > X2 > ... > Xd}.  The first variate @altmath{x_1, X1}
is picked uniformly between 0 and 1, the next variate @altmath{x_2, X2}
between 0 and @altmath{x_1, X1}, and so on.  This distribution is so
non-uniform that it should essentially @emph{never} be detected as
uniform.

@item normal
The multivariate normal distribution with mean at the origin and
identity covariance matrix.
@end table

@item -m @var{mst}
@itemx --mst=@var{mst}
Selects the minimum spanning tree algorithm to use, one of:

@table @code
@item prim-bin
Prim's algorithm with a binary heap priority queue.

@item prim-fib
Prim's algorithm with a Fibonacci heap priority queue.
@end table

@item -p @var{proximity}
@itemx --prox=@var{proximity}
The default behavior is to select all of the test points from a single
distribution of the type specified.  When this option is specified along
with a positive value for @var{proximity}, the data points are instead
divided into two equal-size groups, the means of which are displaced
apart a distance of @var{proximity}.

@item -r @var{runs}
@itemx --runs=@var{runs}
Perform the specified tests on @var{runs} sets of data.  The default is
16.

@item -s @var{seed}
@itemx --seed=@var{seed}
Use @var{seed} as the initial random number seed for the first run.
The seed is incremented for each succeeding run.  The default is based
on the current time.

@item -h
@itemx --help
Outputs a brief help screen and exits.

@item -V
@itemx --version
Outputs version and licensing information and exits.
@end table

These features will be used extensively to produce data for the
discussion below.

@ignore
@iftex
Here is a sample of @code{hull-test}'s output.  Each circle is one of
the initial points, each cross a generated point:

@center @image{hull}
@end iftex
@end ignore

@node Testing Results, Sources of Error, Programs for Testing, Discussion
@section Testing Results

@tex
\global\def\hf{\hskip 1pt plus1fill}
@end tex
@ifnottex
@macro hf {}
@end macro
@end ifnottex

The test programs above can be used to analyze the correctness and
performance of the uniformity testing library.  This section presents
results obtained using them along with a discussion of these results'
meaning.

@menu
* Correctness::                 
* Normal Distributions::        
* Testing Separation::          
@end menu

@node Correctness, Normal Distributions, Testing Results, Testing Results
@subsection Correctness

Correctness of the testing algorithm can be tested both when the
sampling window is known and when it is unknown.  The ordinary
@code{unf-test} program can be used for the latter; the former can be
tested the same way by temporarily commenting out the line
@display
|if (unf_inside_hull (p, n, d, y, tmp))| 
@end display
@noindent 
in @<Generate points for uniformity test@> and recompiling.@footnote{The
proper way to do this would be to avoid use of |unf_test()|, using the
low-level functions prototyped in @(uniformity.h@> instead.  This is not
done in @code{unf-test} because one of its purposes is to ensure that
|unf_test()| works.}

Tables below show the results of uniformity testing for cells with
various dimensions |d| and point counts |n|.  Each cell is of the form
@math{(R(0.05), R(0.02))}, where @math{R(s)} indicates the percentage of uniform
distributions wrongly detected as non-uniform at the given level |s|.
Each entry corresponds to 10,000 runs.

@menu
* Known Sampling Window::       
* Unknown Sampling Window::     
@end menu

@node Known Sampling Window, Unknown Sampling Window, Correctness, Correctness
@subsubsection Known Sampling Window

This table shows the results of uniformity testing for uniform samples
when the sampling window is known:

@multitable @columnfractions .175 .075 .075 .15 .15 .15
@item @tab @tab @tab @tab @hf{}@w{    }|d|@hf{}
@item @tab @tab @tab @hf{}@w{    }2@hf{} @tab @hf{}@w{    }5@hf{} @tab @hf{}@w{    }10@hf{}
@item @tab @tab 50 @tab @hf{}(4.3,1.5)@hf{} @tab @hf{}(4.2,1.7)@hf{} @tab @hf{}(4.2,1.6)@hf{}
@item @tab |n| @tab 100 @tab @hf{}(5.0,1.9)@hf{} @tab @hf{}(5.2,2.0)@hf{} @tab @hf{}(4.6,2.0)@hf{}
@item @tab @tab 200 @tab @hf{}(4.8,2.0)@hf{} @tab @hf{}(4.5,2.1)@hf{} @tab @hf{}(5.0,1.9)@hf{}
@end multitable

The expected values for R(0.05) and R(0.02) are 5 and 2, respectively.
Considering each uniformity test R(s) as a binomial trial with
probability of success |p = s| and selecting a significance level
@altmath{\alpha = 0.05, alpha = 0.05}, the allowed intervals to accept
the hypothesis @altmath{p = \widehat{p}, p = p_hat} are @altmath{4.57 <
R(0.05) < 5.43, 4.57 < R(0.05) < 5.43} and @altmath{1.73 < R(0.02) <
2.27, 1.73 < R(0.02) < 2.27}.

For |n = 50|, few of the values in the table are in the range for
acceptance, but for |n = 100| and |n = 200| most of them are.  This
indicates that the algorithm is probably correct.  A later section will
discuss possible sources of error.

In the original paper, all of the values are within the range to allow
acceptance, but only 100 runs were made per entry, instead of 10,000,
presumably because of time constraints.  When only 100 runs are made, it
is easy to accidentally get results that are all within the acceptable
range.

@references
@bibref{Smith 1984}, table I;
@bibref{Zwillinger 1996}, sections 7.9.1 and 7.12.2.

@node Unknown Sampling Window,  , Known Sampling Window, Correctness
@subsubsection Unknown Sampling Window

When the sampling window is unknown, the results are like this:

@multitable @columnfractions .175 .075 .075 .15 .15 .15
@item @tab @tab @tab @tab @hf{}@w{    }|d|@hf{}
@item @tab @tab @tab @hf{}@w{    }2@hf{} @tab @hf{}@w{    }5@hf{} @tab @hf{}@w{    }10@hf{}
@item @tab @tab 50 @tab @hf{}(4.2,1.6)@hf{} @tab @hf{}(2.3,0.9)@hf{} @tab @hf{}(0.5,0.1)@hf{}
@item @tab |n| @tab 100 @tab @hf{}(5.3,1.8)@hf{} @tab @hf{}(3.0,1.0)@hf{} @tab @hf{}(0.7,0.2)@hf{}
@item @tab @tab 200 @tab @hf{}(4.6,1.9)@hf{} @tab @hf{}(3.3,1.1)@hf{} @tab @hf{}(0.5,0.2)@hf{}
@end multitable

@noindent
For small |d|, entries are in the same general range as for the known
sampling window data, and as |d| increases, they become more
conservative.  This is the expected behavior according to the
references.

@references
@bibref{Smith 1984}, table IV.

@node Normal Distributions, Testing Separation, Correctness, Testing Results
@subsection Normal Distributions

We can run the uniformity test on points selected from a normal
distribution.  If this is done for several values of |n| with |d = 5|
and a desired 5% false detection rate, this graph results:

@ifnotinfo
@center @image{normal, 3in}
@end ifnotinfo
@ifinfo
@center (Refer to high-quality output for graph.)
@end ifinfo

In words, for small |n|, uniform and normal distributions are almost
indistinguishable.  As |n| increases, the distribution is detected more
and more often as non-uniform, until the time that |n = 200| is reached,
at which it is essentially always found to be non-uniform.

@node Testing Separation,  , Normal Distributions, Testing Results
@subsection Testing Separation

An additional interesting test is to divide the test points into two
sets and separate their means by increasing distances.  The graphs below
show the results for (a) normal distributions and (b) uniform
distributions, each with a total of |n = 50| points.  Note that the
graphs are nearly identical.

@ifnotinfo
@center (a)@image{sepnormal, 2.5in} (b)@image{sepuniform, 2.5in}
@end ifnotinfo
@ifinfo
@center (Refer to high-quality output for graph.)
@end ifinfo

@node Sources of Error, Performance, Testing Results, Discussion
@section Sources of Error

Based on results above, the code implementing the uniformity testing
algorithm seems to work properly.  This section discusses possible
sources of error that were considered during design and implementation
but have not already been discussed.

@menu
* Bugs in MST Implementation::  
* Bugs in Convex Hull Insideness Test::  
* Imprecise Arithmetic::        
@end menu

@node Bugs in MST Implementation, Bugs in Convex Hull Insideness Test, Sources of Error, Sources of Error
@subsection Bugs in MST Implementation

There are two programs to test the minimum spanning tree algorithms:
@code{pq-test} for the priority queues and @code{mst-test} for the
spanning tree itself.

Testing the queue implementations for obvious flaws by running the
@code{pq-test} program with several sets of parameters does not reveal
any problems in either priority queue implementation.  This does not
mean that none exist, but it is a good sign.

Running @code{mst-test} shows that it produces correct minimum spanning
trees on test 
@ifnotinfo
data.  Minimum spanning tree (a) below was produced with the binary heap
priority queue, (b) with the Fibonacci heap priority queue:

@center (a)@image{mst-bin, 2in} (b)@image{mst-fib, 2in}
@end ifnotinfo
@ifinfo
data. (Refer to a high-quality output version of this report for
examples of @code{mst-test} output.)
@end ifinfo

@node Bugs in Convex Hull Insideness Test, Imprecise Arithmetic, Bugs in MST Implementation, Sources of Error
@subsection Bugs in Convex Hull Insideness Test

The test whether a point is inside the approximate convex hull of the
set of points is suspicious from the beginning because of the way that
the equation for the test is written in both the dissertation and the
paper describing the algorithm, using a subscript `2' in a confusing
way.  Presumably a superscript `2' was intended; it was interpreted this
way in the implementation.
@iftex
The original form given for this equation was:
@tex
$$\widehat{n}^* = N^{-1} \sum_{i=1}^N (Xi - Y)/(\Vert Xi - Y\Vert_2)^{K+1}.$$
@end tex
@end iftex
A search of the literature did not turn up use of any similar tests, so
no comparisons could be made.  Several possible variants of this test
were tried, but none produced better results.

If the @code{hull-test} program is used to visually test the convex hull
estimation, the results are 
@ifnotinfo
disappointing:

@center @image{hull, 2in}
@end ifnotinfo
@ifinfo
disappointing.  (Refer to the a high-quality output version of this
report for an example.)
@end ifinfo
The original points are marked with circles, the ones generated within
the ``approximate convex hull'' with crosses.  Notice how far away from
the triangular border some of the generated points fall.  None of
generated points in a similar illustration in the original paper fall as
far away as these.

@references
@bibref{Smith 1984}, figure 3.

@node Imprecise Arithmetic,  , Bugs in Convex Hull Insideness Test, Sources of Error
@subsection Imprecise Arithmetic

The uniformity testing library uses |double| for all of its
floating-point calculations.  It is possible that additional precision
might increase accuracy in calculations.

To experiment with this idea, all |double| (64-bit IEEE 754) quantities
were replaced by |long double| (80-bit IEEE 754) quantities.  In
addition, sums were performed both using Kahan summation and in
least-to-greatest order (by performing a sort operation).

There was no noticeable change in behavior from these increases in
precision.  In conclusion, for the current data sets, arithmetic
precision is not a limiting factor.

@references
@bibref{Goldberg 1991}.

@node Performance, Future Work, Sources of Error, Discussion
@section Performance

The two significant steps in the uniformity test whose performance
depend on |n| and |d| are the generation of random points and
calculation of the minimum spanning tree.  The first of these has
asymptotic performance O(@altmath{n^2, n**2}), the latter O(@altmath{n^2
\log n, n**2 lg n}) for Prim's algorithm with a binary heap or
O(@altmath{n^2, n**2}) with a Fibonacci heap.

When @code{unf-test} is compiled with GCC 2.95.4 prerelease 20010319
using options @code{-O3 -DNDEBUG} and run on a Pentium II/233 MHz, tests
with various |n| and |d| using the binary heap priority queue, run 50
times each, take the times shown in the table below, in seconds:

@multitable @columnfractions .175 .075 .075 .15 .15 .15
@item @tab @tab @tab @tab @hf{}|d|
@item @tab @tab @tab @hf{}2 @tab @hf{}5 @tab @hf{}10
@item @tab @tab 50 @tab @hf{}0.3 @tab @hf{}0.4 @tab @hf{}0.7
@item @tab |n| @tab 100 @tab @hf{}1.2 @tab @hf{}1.4 @tab @hf{}2.2
@item @tab @tab 200 @tab @hf{}4.3 @tab @hf{}5.1 @tab @hf{}8.0
@item @tab @tab 400 @tab @hf{}17.2 @tab @hf{}20.5 @tab @hf{}30.2
@item @tab @tab 800 @tab @hf{}69.1 @tab @hf{}85.2 @tab @hf{}116.6
@end multitable

@noindent
Note that doubling |n| causes the time to quadruple, almost exactly, so
the practical performance of the code for |n| in these ranges is
O(@altmath{n^2, n**2}) under these conditions.

When a Fibonacci heap is used instead, the times @emph{increase}
slightly:

@multitable @columnfractions .175 .075 .075 .15 .15 .15
@item @tab @tab @tab @tab @hf{}|d|
@item @tab @tab @tab @hf{}2 @tab @hf{}5 @tab @hf{}10
@item @tab @tab 50 @tab @hf{}0.4 @tab @hf{}0.4 @tab @hf{}0.7
@item @tab |n| @tab 100 @tab @hf{}1.2 @tab @hf{}1.4 @tab @hf{}2.3
@item @tab @tab 200 @tab @hf{}4.6 @tab @hf{}5.5 @tab @hf{}8.2
@item @tab @tab 400 @tab @hf{}18.4 @tab @hf{}21.7 @tab @hf{}31.1
@item @tab @tab 800 @tab @hf{}72.5 @tab @hf{}86.8 @tab @hf{}119.6
@end multitable

@noindent
This effect is presumably due to the increased complexity of a Fibonacci
heap.  According to @bibref{Cormen 1990}, chapter 21, this is not
unexpected:

@quotation
From a practical point of view, however, the constant factors and
programming complexity of Fibonacci heaps make them less desirable than
ordinary binary (or |k|-ary) heaps for most applications.  Thus,
Fibonacci heaps are predominantly of theoretical interest.
@end quotation

@node Future Work,  , Performance, Discussion
@section Future Work

Currently the code for the uniformity testing library is very general,
because it was written with little idea of how it was going to be used.
Specializing it for particular situations may be one task for future
work.  For instance, some applications may prefer the use of a bounding
hypersphere instead of the provided hyper-rectangle implementation.

There is much potential for optimization.  Profiling shows that
generation and testing of points and the minimum spanning tree
computation take roughly equal amounts of time, so improvement in either
one would improve performance.  A few minutes work on the implementation
of the Prim MST in particular might help.  Some experimentation shows
that a 10% improvement there, by modifying its interface to the priority
queue, is fairly simple to achieve, and a larger improvement might not
be much more difficult.  Replacement of Prim's algorithm by one of
Bentley's algorithms might reap much greater benefits.

This document could also use some elaboration for the benefit of less
knowledgeable readers.  For instance, a better explanation of Prim's
algorithm for constructing MST and for Fibonacci heap algorithms would
be helpful.
